@@ -1,5 +1,8 @@
 /*
- * Copyright 1990 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1990 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1996 - 2000, Marek Michałkiewicz
+ * Copyright (c) 2001 - 2006, Tomasz Kłoczko
+ * Copyright (c) 2007 - 2008, Nicolas François
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -10,34 +13,39 @@
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. Neither the name of Julianne F. Haugh nor the names of its contributors
- *    may be used to endorse or promote products derived from this software
- *    without specific prior written permission.
+ * 3. The name of the copyright holders or contributors may not be used to
+ *    endorse or promote products derived from this software without
+ *    specific prior written permission.
  *
- * THIS SOFTWARE IS PROVIDED BY JULIE HAUGH AND CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED. IN NO EVENT SHALL JULIE HAUGH OR CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
+ * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 #include <config.h>
 
-#include "rcsid.h"
-RCSID (PKG_VER "$Id: newgrp.c,v 1.28 2005/05/19 11:28:27 kloczek Exp $")
-#include <stdio.h>
+#ident "$Id: newgrp.c 1984 2008-05-18 16:38:13Z nekral-guest $"
+
 #include <errno.h>
 #include <grp.h>
 #include <pwd.h>
-#include "prototypes.h"
+#include <stdio.h>
 #include "defines.h"
 #include "getdef.h"
+#include "prototypes.h"
+#include "exitcodes.h"
+/*
+ * Global variables
+ */
+extern char **newenvp;
 extern char **environ;
 
 #ifdef HAVE_SETGROUPS
@@ -48,47 +56,308 @@ static GETGROUPS_T *grouplist;
 static char *Prog;
 static int is_newgrp;
 
+#ifdef WITH_AUDIT
+char audit_buf[80];
+#endif
+
 /* local function prototypes */
 static void usage (void);
+static void check_perms (const struct group *grp,
+                         struct passwd *pwd,
+                         const char *groupname);
+static void syslog_sg (const char *name, const char *group);
 
 /*
  * usage - print command usage message
  */
-
 static void usage (void)
 {
 	if (is_newgrp)
-		fprintf (stderr, _("Usage: newgrp [-] [group]\n"));
+		fputs (_("Usage: newgrp [-] [group]\n"), stderr);
 	else
-		fprintf (stderr, _("Usage: sg group [[-c] command]\n"));
+		fputs (_("Usage: sg group [[-c] command]\n"), stderr);
 }
 
+/*
+ * find_matching_group - search all groups of a given group id for
+ *                       membership of a given username
+ */
+static struct group *find_matching_group (const char *name, gid_t gid)
+{
+	struct group *gr;
+	char **look;
+	int notfound = 1;
+
+	setgrent ();
+	while ((gr = getgrent ()) != NULL) {
+		if (gr->gr_gid != gid) {
+			continue;
+		}
+
+		/*
+		 * A group with matching GID was found.
+		 * Test for membership of 'name'.
+		 */
+		look = gr->gr_mem;
+		while (*look && (notfound = strcmp (*look++, name)));
+		if (!notfound)
+			break;
+	}
+	endgrent ();
+	return gr;
+}
+
+/*
+ * check_perms - check if the user is allowed to switch to this group
+ *
+ *	If needed, the user will be authenticated.
+ *
+ *	It will not return if the user could not be authenticated.
+ */
+static void check_perms (const struct group *grp,
+                         struct passwd *pwd,
+                         const char *groupname)
+{
+	int needspasswd = 0;
+	struct spwd *spwd;
+	char *cp;
+	const char *cpasswd;
+
+	/*
+	 * see if she is a member of this group (i.e. in the list of
+	 * members of the group, or if the group is her primary group).
+	 *
+	 * If she isn't a member, she needs to provide the group password.
+	 * If there is no group password, she will be denied access
+	 * anyway.
+	 *
+	 */
+	if (grp->gr_gid != pwd->pw_gid && !is_on_list (grp->gr_mem, pwd->pw_name))
+		needspasswd = 1;
+
+	/*
+	 * If she does not have either a shadowed password, or a regular
+	 * password, and the group has a password, she needs to give the
+	 * group password.
+	 */
+	spwd = xgetspnam (pwd->pw_name);
+	if (NULL != spwd)
+		pwd->pw_passwd = spwd->sp_pwdp;
+
+	if (pwd->pw_passwd[0] == '\0' && grp->gr_passwd[0])
+		needspasswd = 1;
+
+	/*
+	 * Now I see about letting her into the group she requested. If she
+	 * is the root user, I'll let her in without having to prompt for
+	 * the password. Otherwise I ask for a password if she flunked one
+	 * of the tests above.
+	 */
+	if (getuid () != 0 && needspasswd) {
+		/*
+		 * get the password from her, and set the salt for
+		 * the decryption from the group file.
+		 */
+		cp = getpass (_("Password: "));
+		if (NULL == cp)
+			goto failure;
+
+		/*
+		 * encrypt the key she gave us using the salt from the
+		 * password in the group file. The result of this encryption
+		 * must match the previously encrypted value in the file.
+		 */
+		cpasswd = pw_encrypt (cp, grp->gr_passwd);
+		strzero (cp);
+
+		if (grp->gr_passwd[0] == '\0' ||
+		    strcmp (cpasswd, grp->gr_passwd) != 0) {
+#ifdef WITH_AUDIT
+			snprintf (audit_buf, sizeof(audit_buf),
+			          "authentication new-gid=%d", grp->gr_gid);
+			audit_logger (AUDIT_GRP_AUTH, Prog,
+			              audit_buf, NULL, getuid (), 0);
+#endif
+			SYSLOG ((LOG_INFO,
+				 "Invalid password for group `%s' from `%s'",
+				 groupname, pwd->pw_name));
+			sleep (1);
+			fputs (_("Invalid password.\n"), stderr);
+			goto failure;
+		}
+#ifdef WITH_AUDIT
+		snprintf (audit_buf, sizeof(audit_buf),
+		          "authentication new-gid=%d", grp->gr_gid);
+		audit_logger (AUDIT_GRP_AUTH, Prog,
+		              audit_buf, NULL, getuid (), 1);
+#endif
+	}
+
+	return;
+
+failure:
+	/* The closelog is probably unnecessary, but it does no
+	 * harm.  -- JWP
+	 */
+	closelog ();
+#ifdef WITH_AUDIT
+	if (groupname) {
+		snprintf (audit_buf, sizeof(audit_buf),
+		          "changing new-group=%s", groupname);
+		audit_logger (AUDIT_CHGRP_ID, Prog,
+		              audit_buf, NULL, getuid (), 0);
+	} else {
+		audit_logger (AUDIT_CHGRP_ID, Prog, "changing",
+		              NULL, getuid (), 0);
+	}
+#endif
+	exit (1);
+}
+
+#ifdef USE_SYSLOG
+/*
+ * syslog_sg - log the change of group to syslog
+ *
+ *	The loggout will also be logged when the user will quit the
+ *	sg/newgrp session.
+ */
+static void syslog_sg (const char *name, const char *group)
+{
+	const char *loginname = getlogin ();
+	const char *tty = ttyname (0);
+
+	if (loginname != NULL)
+		loginname = xstrdup (loginname);
+	if (tty != NULL)
+		tty = xstrdup (tty);
+
+	if (loginname == NULL)
+		loginname = "???";
+	if (tty == NULL)
+		tty = "???";
+	else if (strncmp (tty, "/dev/", 5) == 0)
+		tty += 5;
+	SYSLOG ((LOG_INFO,
+		 "user `%s' (login `%s' on %s) switched to group `%s'",
+		 name, loginname, tty, group));
+#ifdef USE_PAM
+	/*
+	 * We want to fork and exec the new shell in the child, leaving the
+	 * parent waiting to log the session close.
+	 *
+	 * The parent must ignore signals generated from the console
+	 * (SIGINT, SIGQUIT, SIGHUP) which might make the parent terminate
+	 * before its child. When bash is exec'ed as the subshell, it
+	 * generates a new process group id for itself, and consequently
+	 * only SIGHUP, which is sent to all process groups in the session,
+	 * can reach the parent. However, since arbitrary programs can be
+	 * specified as login shells, there is no such guarantee in general.
+	 * For the same reason, we must also ignore stop signals generated
+	 * from the console (SIGTSTP, SIGTTIN, and SIGTTOU) in order to
+	 * avoid any possibility of the parent being stopped when it
+	 * receives SIGCHLD from the terminating subshell.  -- JWP
+	 */
+	{
+		pid_t child, pid;
+
+		signal (SIGINT, SIG_IGN);
+		signal (SIGQUIT, SIG_IGN);
+		signal (SIGHUP, SIG_IGN);
+		signal (SIGTSTP, SIG_IGN);
+		signal (SIGTTIN, SIG_IGN);
+		signal (SIGTTOU, SIG_IGN);
+		child = fork ();
+		if (child < 0) {
+			/* error in fork() */
+			fprintf (stderr, _("%s: failure forking: %s\n"),
+				 is_newgrp ? "newgrp" : "sg", strerror (errno));
+#ifdef WITH_AUDIT
+			if (group) {
+				snprintf (audit_buf, sizeof(audit_buf),
+				          "changing new-group=%s", group);
+				audit_logger (AUDIT_CHGRP_ID, Prog,
+				              audit_buf, NULL, getuid (), 0);
+			} else {
+				audit_logger (AUDIT_CHGRP_ID, Prog, "changing",
+				              NULL, getuid (), 0);
+			}
+#endif
+			exit (1);
+		} else if (child) {
+			/* parent - wait for child to finish, then log session close */
+			int cst = 0;
+			gid_t gid = getgid();
+			struct group *grp = getgrgid (gid);
+
+			do {
+				errno = 0;
+				pid = waitpid (child, &cst, WUNTRACED);
+				if (pid == child && WIFSTOPPED (cst)) {
+					/* stop when child stops */
+					kill (getpid (), WSTOPSIG(cst));
+					/* wake child when resumed */
+					kill (child, SIGCONT);
+				}
+			} while ((pid == child && WIFSTOPPED (cst)) ||
+				 (pid != child && errno == EINTR));
+			/* local, no need for xgetgrgid */
+			if (NULL != grp) {
+				SYSLOG ((LOG_INFO,
+				         "user `%s' (login `%s' on %s) returned to group `%s'",
+				         name, loginname, tty, grp->gr_name));
+			} else {
+				SYSLOG ((LOG_INFO,
+				         "user `%s' (login `%s' on %s) returned to group `%d'",
+				         name, loginname, tty, gid));
+				/* Either the user's passwd entry has a
+				 * GID that does not match with any group,
+				 * or the group was deleted while the user
+				 * was in a newgrp session.*/
+				SYSLOG ((LOG_WARN,
+				         "unknown GID `%u' used by user `%s'",
+				         gid, name));
+			}
+			closelog ();
+			exit (0);
+		}
+
+		/* child - restore signals to their default state */
+		signal (SIGINT, SIG_DFL);
+		signal (SIGQUIT, SIG_DFL);
+		signal (SIGHUP, SIG_DFL);
+		signal (SIGTSTP, SIG_DFL);
+		signal (SIGTTIN, SIG_DFL);
+		signal (SIGTTOU, SIG_DFL);
+	}
+#endif				/* USE_PAM */
+}
+#endif				/* USE_SYSLOG */
+
 /*
  * newgrp - change the invokers current real and effective group id
  */
-
 int main (int argc, char **argv)
 {
 	int initflag = 0;
-	int needspasswd = 0;
 	int i;
 	int cflag = 0;
+	int err = 0;
 	gid_t gid;
 	char *cp;
-	const char *cpasswd, *name, *prog;
+	const char *name, *prog;
 	char *group = NULL;
 	char *command = NULL;
 	char **envp = environ;
 	struct passwd *pwd;
 	struct group *grp;
 
-#ifdef SHADOWPWD
-	struct spwd *spwd;
-#endif
 #ifdef SHADOWGRP
 	struct sgrp *sgrp;
 #endif
 
+#ifdef WITH_AUDIT
+	audit_help_open ();
+#endif
 	setlocale (LC_ALL, "");
 	bindtextdomain (PACKAGE, LOCALEDIR);
 	textdomain (PACKAGE);
@@ -116,7 +385,6 @@ int main (int argc, char **argv)
 	 * but we do not need to restore the previous process persona and we
 	 * don't need to re-exec anything.  -- JWP
 	 */
-
 	Prog = Basename (argv[0]);
 	is_newgrp = (strcmp (Prog, "newgrp") == 0);
 	OPENLOG (is_newgrp ? "newgrp" : "sg");
@@ -128,8 +396,12 @@ int main (int argc, char **argv)
 
 	pwd = get_my_pwent ();
 	if (!pwd) {
-		fprintf (stderr, _("unknown uid: %u\n"), getuid ());
-		SYSLOG ((LOG_WARN, "unknown uid %u", getuid ()));
+		fprintf (stderr, _("unknown UID: %u\n"), getuid ());
+#ifdef WITH_AUDIT
+		audit_logger (AUDIT_CHGRP_ID, Prog, "changing", NULL,
+			      getuid (), 0);
+#endif
+		SYSLOG ((LOG_WARN, "unknown UID %u", getuid ()));
 		closelog ();
 		exit (1);
 	}
@@ -152,7 +424,6 @@ int main (int argc, char **argv)
 	 *      sg [-]
 	 *      sg [-] groupid [[-c command]
 	 */
-
 	if (argc > 0 && (!strcmp (argv[0], "-") || !strcmp (argv[0], "-l"))) {
 		argc--;
 		argv++;
@@ -163,7 +434,6 @@ int main (int argc, char **argv)
 		 * Do the command line for everything that is
 		 * not "newgrp".
 		 */
-
 		if (argc > 0 && argv[0][0] != '-') {
 			group = argv[0];
 			argc--;
@@ -180,7 +450,6 @@ int main (int argc, char **argv)
 			 * "sg group -c command" (as in the man page) or
 			 * "sg group command" (as in the usage message).
 			 */
-
 			if (argc > 1 && strcmp (argv[0], "-c") == 0)
 				command = argv[1];
 			else
@@ -188,19 +457,16 @@ int main (int argc, char **argv)
 			cflag++;
 		}
 	} else {
-
 		/*
 		 * Do the command line for "newgrp". It's just making sure
 		 * there aren't any flags and getting the new group name.
 		 */
-
 		if (argc > 0 && argv[0][0] == '-') {
 			usage ();
 			goto failure;
 		} else if (argv[0] != (char *) 0) {
 			group = argv[0];
 		} else {
-
 			/*
 			 * get the group file entry for her login group id. 
 			 * the entry must exist, simply to be annoying.
@@ -208,11 +474,11 @@ int main (int argc, char **argv)
 			 * Perhaps in the past, but the default behavior now depends on the
 			 * group entry, so it had better exist.  -- JWP
 			 */
-
-			if (!(grp = getgrgid (pwd->pw_gid))) {
-				fprintf (stderr, _("unknown gid: %lu\n"),
+			grp = xgetgrgid (pwd->pw_gid);
+			if (NULL == grp) {
+				fprintf (stderr, _("unknown GID: %lu\n"),
 					 (unsigned long) pwd->pw_gid);
-				SYSLOG ((LOG_CRIT, "unknown gid: %lu",
+				SYSLOG ((LOG_CRIT, "unknown GID: %lu",
 					 (unsigned long) pwd->pw_gid));
 				goto failure;
 			} else
@@ -227,7 +493,6 @@ int main (int argc, char **argv)
 	 * nasty message but at least your real and effective group id's are
 	 * set.
 	 */
-
 	/* don't use getgroups(0, 0) - it doesn't work on some systems */
 	i = 16;
 	for (;;) {
@@ -241,6 +506,17 @@ int main (int argc, char **argv)
 	}
 	if (ngroups < 0) {
 		perror ("getgroups");
+#ifdef WITH_AUDIT
+		if (group) {
+			snprintf (audit_buf, sizeof(audit_buf),
+			          "changing new-group=%s", group);
+			audit_logger (AUDIT_CHGRP_ID, Prog,
+			              audit_buf, NULL, getuid (), 0);
+		} else {
+			audit_logger (AUDIT_CHGRP_ID, Prog,
+			              "changing", NULL, getuid (), 0);
+		}
+#endif
 		exit (1);
 	}
 #endif				/* HAVE_SETGROUPS */
@@ -252,14 +528,14 @@ int main (int argc, char **argv)
 	 * set to the value from her password file entry.  
 	 *
 	 * If run as newgrp, or as sg with no command, this process exec's
-	 * an interactive subshell with the effective gid of the new group. 
+	 * an interactive subshell with the effective GID of the new group. 
 	 * If run as sg with a command, that command is exec'ed in this
 	 * subshell. When this process terminates, either because the user
 	 * exits, or the command completes, the parent of this process
-	 * resumes with the current gid.
+	 * resumes with the current GID.
 	 *
 	 * If a group is explicitly specified on the command line, the
-	 * interactive shell or command is run with that effective gid. 
+	 * interactive shell or command is run with that effective GID. 
 	 * Access will be denied if no entry for that group can be found in
 	 * /etc/group. If the current user name appears in the members list
 	 * for that group, access will be granted immediately; if not, the
@@ -267,9 +543,9 @@ int main (int argc, char **argv)
 	 * password response is incorrect, if the specified group does not
 	 * have a password, or if that group has been locked by gpasswd -R,
 	 * access will be denied. This is true even if the group specified
-	 * has the user's login gid (as shown in /etc/passwd). If no group
+	 * has the user's login GID (as shown in /etc/passwd). If no group
 	 * is explicitly specified on the command line, the effect is
-	 * exactly the same as if a group name matching the user's login gid
+	 * exactly the same as if a group name matching the user's login GID
 	 * had been explicitly specified. Root, however, is never
 	 * challenged for passwords, and is always allowed access.
 	 *
@@ -285,171 +561,48 @@ int main (int argc, char **argv)
 	 * including the user's name in the member list of the user's login
 	 * group.  -- JWP
 	 */
-
-	if (!(grp = getgrnam (group))) {
+	grp = getgrnam (group); /* local, no need for xgetgrnam */
+	if (NULL == grp) {
 		fprintf (stderr, _("unknown group: %s\n"), group);
 		goto failure;
 	}
+
+	/*
+	 * For splitted groups (due to limitations of NIS), check all 
+	 * groups of the same GID like the requested group for
+	 * membership of the current user.
+	 */
+	grp = find_matching_group (name, grp->gr_gid);
+	if (!grp) {
+		/*
+		 * No matching group found. As we already know that
+		 * the group exists, this happens only in the case
+		 * of a requested group where the user is not member.
+		 *
+		 * Re-read the group entry for further processing.
+		 */
+		grp = xgetgrnam (group);
+	}
 #ifdef SHADOWGRP
-	if ((sgrp = getsgnam (group))) {
+	sgrp = getsgnam (group);
+	if (NULL != sgrp) {
 		grp->gr_passwd = sgrp->sg_passwd;
 		grp->gr_mem = sgrp->sg_mem;
 	}
 #endif
 
 	/*
-	 * see if she is a member of this group. If she isn't a member, she
-	 * needs to provide the group password. If there is no group
-	 * password, she will be denied access anyway.
-	 *
+	 * Check if the user is allowed to access this group.
 	 */
-
-	if (!is_on_list (grp->gr_mem, name))
-		needspasswd = 1;
-
-	/*
-	 * If she does not have either a shadowed password, or a regular
-	 * password, and the group has a password, she needs to give the
-	 * group password.
-	 */
-
-#ifdef	SHADOWPWD
-	if ((spwd = getspnam (name)))
-		pwd->pw_passwd = spwd->sp_pwdp;
-#endif
-
-	if (pwd->pw_passwd[0] == '\0' && grp->gr_passwd[0])
-		needspasswd = 1;
-
-	/*
-	 * Now i see about letting her into the group she requested. If she
-	 * is the root user, i'll let her in without having to prompt for
-	 * the password. Otherwise i ask for a password if she flunked one
-	 * of the tests above.
-	 *
-	 * Note that she now has to provide the password to her own group,
-	 * unless she is listed as a member.  -- JWP
-	 */
-
-	if (getuid () != 0 && needspasswd) {
-
-		/*
-		 * get the password from her, and set the salt for
-		 * the decryption from the group file.
-		 */
-
-		if (!(cp = getpass (_("Password: "))))
-			goto failure;
-
-		/*
-		 * encrypt the key she gave us using the salt from the
-		 * password in the group file. The result of this encryption
-		 * must match the previously encrypted value in the file.
-		 */
-
-		cpasswd = pw_encrypt (cp, grp->gr_passwd);
-		strzero (cp);
-
-		if (grp->gr_passwd[0] == '\0') {
-			/*
-			 * there is no password, print out "Sorry" and give up
-			 */
-			sleep (1);
-			fputs (_("Sorry.\n"), stderr);
-			goto failure;
-		}
-
-		if (strcmp (cpasswd, grp->gr_passwd) != 0) {
-			SYSLOG ((LOG_INFO,
-				 "Invalid password for group `%s' from `%s'",
-				 group, name));
-			sleep (1);
-			fputs (_("Sorry.\n"), stderr);
-			goto failure;
-		}
-	}
+	check_perms (grp, pwd, group);
 
 	/*
 	 * all successful validations pass through this point. The group id
 	 * will be set, and the group added to the concurrent groupset.
 	 */
-
 #ifdef	USE_SYSLOG
-	if (getdef_bool ("SYSLOG_SG_ENAB"))
-		SYSLOG ((LOG_INFO, "user `%s' switched to group `%s'",
-			 name, group));
 	if (getdef_bool ("SYSLOG_SG_ENAB")) {
-		char *loginname = getlogin ();
-		char *tty = ttyname (0);
-
-		if (loginname != NULL)
-			loginname = xstrdup (loginname);
-		if (tty != NULL)
-			tty = xstrdup (tty);
-
-		if (loginname == NULL)
-			loginname = "???";
-		if (tty == NULL)
-			tty = "???";
-		else if (strncmp (tty, "/dev/", 5) == 0)
-			tty += 5;
-		SYSLOG ((LOG_INFO,
-			 "user `%s' (login `%s' on %s) switched to group `%s'",
-			 name, loginname, tty, group));
-#ifdef USE_PAM
-		if (getdef_bool ("CLOSE_SESSIONS")) {
-			/*
-			 * We want to fork and exec the new shell in the child, leaving the
-			 * parent waiting to log the session close.
-			 *
-			 * The parent must ignore signals generated from the console
-			 * (SIGINT, SIGQUIT, SIGHUP) which might make the parent terminate
-			 * before its child. When bash is exec'ed as the subshell, it
-			 * generates a new process group id for itself, and consequently
-			 * only SIGHUP, which is sent to all process groups in the session,
-			 * can reach the parent. However, since arbitrary programs can be
-			 * specified as login shells, there is no such guarantee in general.
-			 * For the same reason, we must also ignore stop signals generated
-			 * from the console (SIGTSTP, SIGTTIN, and SIGTTOU) in order to
-			 * avoid any possibility of the parent being stopped when it
-			 * receives SIGCHLD from the terminating subshell.  -- JWP
-			 */
-			pid_t child, pid;
-
-			signal (SIGINT, SIG_IGN);
-			signal (SIGQUIT, SIG_IGN);
-			signal (SIGHUP, SIG_IGN);
-			signal (SIGTSTP, SIG_IGN);
-			signal (SIGTTIN, SIG_IGN);
-			signal (SIGTTOU, SIG_IGN);
-			child = fork ();
-			if (child < 0) {
-				/* error in fork() */
-				fprintf (stderr, "%s: failure forking: %s",
-					 is_newgrp ? "newgrp" : "sg",
-					 strerror (errno));
-				exit (1);
-			} else if (child) {
-				/* parent - wait for child to finish, then log session close */
-				do {
-					pid = waitpid (child, NULL, 0);
-				} while (pid != child);
-				SYSLOG ((LOG_INFO,
-					 "user `%s' (login `%s' on %s) returned to group `%s'",
-					 name, loginname, tty,
-					 getgrgid (gid)->gr_name));
-				closelog ();
-				exit (0);
-			}
-			/* child - restore signals to their default state */
-			signal (SIGINT, SIG_DFL);
-			signal (SIGQUIT, SIG_DFL);
-			signal (SIGHUP, SIG_DFL);
-			signal (SIGTSTP, SIG_DFL);
-			signal (SIGTTIN, SIG_DFL);
-			signal (SIGTTOU, SIG_DFL);
-		}
-#endif				/* USE_PAM */
+		syslog_sg (name, group);
 	}
 #endif				/* USE_SYSLOG */
 
@@ -462,14 +615,13 @@ int main (int argc, char **argv)
 	 * If the group doesn't fit, i'll complain loudly and skip this
 	 * part.
 	 */
-
 	for (i = 0; i < ngroups; i++) {
 		if (gid == grouplist[i])
 			break;
 	}
 	if (i == ngroups) {
 		if (ngroups >= sysconf (_SC_NGROUPS_MAX)) {
-			fprintf (stderr, _("too many groups\n"));
+			fputs (_("too many groups\n"), stderr);
 		} else {
 			grouplist[ngroups++] = gid;
 			if (setgroups (ngroups, grouplist)) {
@@ -480,16 +632,29 @@ int main (int argc, char **argv)
 #endif
 
 	/*
-	 * Set the effective gid to the new group id and the effective uid
-	 * to the real uid. For root, this also sets the real gid to the
+	 * Set the effective GID to the new group id and the effective UID
+	 * to the real UID. For root, this also sets the real GID to the
 	 * new group id.
 	 */
-
-	if (setgid (gid))
+	if (setgid (gid)) {
 		perror ("setgid");
+#ifdef WITH_AUDIT
+		snprintf (audit_buf, sizeof(audit_buf),
+		          "changing new-gid=%d", gid);
+		audit_logger (AUDIT_CHGRP_ID, Prog,
+		              audit_buf, NULL, getuid (), 0);
+#endif
+		exit (1);
+	}
 
 	if (setuid (getuid ())) {
 		perror ("setuid");
+#ifdef WITH_AUDIT
+		snprintf (audit_buf, sizeof(audit_buf),
+		          "changing new-gid=%d", gid);
+		audit_logger (AUDIT_CHGRP_ID, Prog,
+		              audit_buf, NULL, getuid (), 0);
+#endif
 		exit (1);
 	}
 
@@ -497,17 +662,17 @@ int main (int argc, char **argv)
 	 * See if the "-c" flag was used. If it was, i just create a shell
 	 * command for her using the argument that followed the "-c" flag.
 	 */
-
 	if (cflag) {
 		closelog ();
 		execl ("/bin/sh", "sh", "-c", command, (char *) 0);
-		if (errno == ENOENT) {
-			perror ("/bin/sh");
-			exit (127);
-		} else {
-			perror ("/bin/sh");
-			exit (126);
-		}
+#ifdef WITH_AUDIT
+		snprintf (audit_buf, sizeof(audit_buf),
+		          "changing new-gid=%d", gid);
+		audit_logger (AUDIT_CHGRP_ID, Prog,
+		              audit_buf, NULL, getuid (), 0);
+#endif
+		perror ("/bin/sh");
+		exit (errno == ENOENT ? E_CMD_NOTFOUND : E_CMD_NOEXEC);
 	}
 
 	/*
@@ -527,7 +692,6 @@ int main (int argc, char **argv)
 	 * problem, try using $SHELL as a workaround; also please notify me
 	 * at jparmele@wildbear.com -- JWP
 	 */
-
 	if (!initflag && (cp = getenv ("SHELL")))
 		prog = cp;
 	else if (pwd->pw_shell && pwd->pw_shell[0])
@@ -536,15 +700,12 @@ int main (int argc, char **argv)
 		prog = "/bin/sh";
 
 	/*
-	 * Now i try to find the basename of the login shell. This will
+	 * Now I try to find the basename of the login shell. This will
 	 * become argv[0] of the spawned command.
 	 */
-
 	cp = Basename ((char *) prog);
 
-#ifdef	SHADOWPWD
 	endspent ();
-#endif
 #ifdef	SHADOWGRP
 	endsgent ();
 #endif
@@ -555,7 +716,6 @@ int main (int argc, char **argv)
 	 * Switch back to her home directory if i am doing login
 	 * initialization.
 	 */
-
 	if (initflag) {
 		if (chdir (pwd->pw_dir))
 			perror ("chdir");
@@ -574,12 +734,16 @@ int main (int argc, char **argv)
 			addenv (*envp++, NULL);
 	}
 
+#ifdef WITH_AUDIT
+	snprintf (audit_buf, sizeof(audit_buf), "changing new-gid=%d", gid);
+	audit_logger (AUDIT_CHGRP_ID, Prog, audit_buf, NULL, getuid (), 1);
+#endif
 	/*
 	 * Exec the login shell and go away. We are trying to get back to
 	 * the previous environment which should be the user's login shell.
 	 */
-
-	shell (prog, initflag ? (char *) 0 : cp);
+	err = shell (prog, initflag ? (char *) 0 : cp, newenvp);
+	exit (err == ENOENT ? E_CMD_NOTFOUND : E_CMD_NOEXEC);
 	/* NOTREACHED */
       failure:
 
@@ -593,7 +757,17 @@ int main (int argc, char **argv)
 	 * process. The closelog is probably unnecessary, but it does no
 	 * harm.  -- JWP
 	 */
-
 	closelog ();
+#ifdef WITH_AUDIT
+	if (group) {
+		snprintf (audit_buf, sizeof(audit_buf),
+		          "changing new-group=%s", group);
+		audit_logger (AUDIT_CHGRP_ID, Prog, 
+		              audit_buf, NULL, getuid (), 0);
+	} else {
+		audit_logger (AUDIT_CHGRP_ID, Prog,
+		              "changing", NULL, getuid (), 0);
+	}
+#endif
 	exit (1);
 }
