@@ -1,5 +1,8 @@
 /*
- * Copyright 1989 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1989 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1996 - 2000, Marek Michałkiewicz
+ * Copyright (c) 2000 - 2006, Tomasz Kłoczko
+ * Copyright (c) 2007 - 2008, Nicolas François
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -10,26 +13,47 @@
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. Neither the name of Julianne F. Haugh nor the names of its contributors
- *    may be used to endorse or promote products derived from this software
- *    without specific prior written permission.
+ * 3. The name of the copyright holders or contributors may not be used to
+ *    endorse or promote products derived from this software without
+ *    specific prior written permission.
  *
- * THIS SOFTWARE IS PROVIDED BY JULIE HAUGH AND CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED. IN NO EVENT SHALL JULIE HAUGH OR CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
+ * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
+/* Some parts substantially derived from an ancestor of:
+   su for GNU.  Run a shell with substitute user and group IDs.
+
+   Copyright (C) 1992-2003 Free Software Foundation, Inc.
+
+   This program is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; either version 2, or (at your option)
+   any later version.
+
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software
+   Foundation, Inc., 51 Franklin Street, Fifth Floor,
+   Boston, MA 02110-1301, USA.  */
+
+
 #include <config.h>
 
-#ident "$Id: su.c,v 1.66 2006/02/08 10:52:49 kloczek Exp $"
+#ident "$Id: su.c 1981 2008-05-18 13:41:56Z nekral-guest $"
 
 #include <getopt.h>
 #include <grp.h>
@@ -107,7 +131,7 @@ static int iswheel (const char *username)
 {
 	struct group *grp;
 
-	grp = getgrnam ("wheel");;
+	grp = getgrnam ("wheel"); /* !USE_PAM, no need for xgetgrnam */
 	if (!grp || !grp->gr_mem)
 		return 0;
 	return is_on_list (grp->gr_mem, username);
@@ -140,14 +164,13 @@ static void su_failure (const char *tty)
 			 oldname[0] ? oldname : "???", name[0] ? name : "???"));
 	closelog ();
 #endif
-	puts (_("Sorry."));
 	exit (1);
 }
 
 
 #ifdef USE_PAM
 /* Signal handler for parent process later */
-static void catch_signals (int sig)
+static void catch_signals (unused int sig)
 {
 	++caught;
 }
@@ -166,7 +189,12 @@ static void run_shell (const char *shellstr, char *args[], int doshell,
 
 	child = fork ();
 	if (child == 0) {	/* child shell */
-		pam_end (pamh, PAM_SUCCESS);
+		/*
+		 * PAM_DATA_SILENT is not supported by some modules, and
+		 * there is no strong need to clean up the process space's
+		 * memory since we will either call exec or exit.
+		pam_end (pamh, PAM_SUCCESS | PAM_DATA_SILENT);
+		 */
 
 		if (doshell)
 			(void) shell (shellstr, (char *) args[0], envp);
@@ -210,9 +238,11 @@ static void run_shell (const char *shellstr, char *args[], int doshell,
 
 			pid = waitpid (-1, &status, WUNTRACED);
 
-			if (WIFSTOPPED (status)) {
-				kill (getpid (), SIGSTOP);
-				/* once we get here, we must have resumed */
+			if ((pid != -1) && WIFSTOPPED (status)) {
+				/* The child (shell) was suspended.
+				 * Suspend su. */
+				kill (getpid (), WSTOPSIG(status));
+				/* wake child when resumed */
 				kill (pid, SIGCONT);
 			}
 		} while (WIFSTOPPED (status));
@@ -241,7 +271,9 @@ static void run_shell (const char *shellstr, char *args[], int doshell,
 		exit (-1);
 	}
 
-	exit (WEXITSTATUS (status));
+	exit (WIFEXITED (status)
+	      ? WEXITSTATUS (status)
+	      : WTERMSIG (status) + 128);
 }
 #endif
 
@@ -250,16 +282,17 @@ static void run_shell (const char *shellstr, char *args[], int doshell,
   */
 static void usage (void)
 {
-	fprintf (stderr, _("Usage: su [options] [login]\n"
-			   "\n"
-			   "Options:\n"
-			   "  -c, --command COMMAND		pass COMMAND to the invoked shell\n"
-			   "  -h, --help			display this help message and exit\n"
-			   "  -, -l, --login		make the shell a login shell\n"
-			   "  -m, -p,\n"
-			   "  --preserve-environment	do not reset environment variables, and keep\n"
-			   "				the same shell\n"
-			   "  -s, --shell SHELL		use SHELL instead of the default in passwd\n"));
+	fputs (_("Usage: su [options] [LOGIN]\n"
+	         "\n"
+	         "Options:\n"
+	         "  -c, --command COMMAND         pass COMMAND to the invoked shell\n"
+	         "  -h, --help                    display this help message and exit\n"
+	         "  -, -l, --login                make the shell a login shell\n"
+	         "  -m, -p,\n"
+	         "  --preserve-environment        do not reset environment variables, and\n"
+	         "                                keep the same shell\n"
+	         "  -s, --shell SHELL             use SHELL instead of the default in passwd\n"
+	         "\n"), stderr);
 	exit (E_USAGE);
 }
 
@@ -276,7 +309,6 @@ static void usage (void)
 int main (int argc, char **argv)
 {
 	char *cp;
-	char **envcp;
 	const char *tty = 0;	/* Name of tty SU is run from        */
 	int doshell = 0;
 	int fakelogin = 0;
@@ -287,11 +319,12 @@ int main (int argc, char **argv)
 	char *shellstr = 0, *command = 0;
 
 #ifdef USE_PAM
+	char **envcp;
 	int ret;
 #else				/* !USE_PAM */
 	int err = 0;
 
-	RETSIGTYPE (*oldsig) ();
+	RETSIGTYPE (*oldsig) (int);
 	int is_console = 0;
 
 	struct spwd *spwd = 0;
@@ -337,19 +370,9 @@ int main (int argc, char **argv)
 		};
 
 		while ((c =
-			getopt_long (argc, argv, "-c:hlmps:", long_options,
+			getopt_long (argc, argv, "c:hlmps:", long_options,
 				     &option_index)) != -1) {
 			switch (c) {
-			case 1:
-				/* this is not an su option */
-				/* The next arguments are either '-', the
-				 * target name, or arguments to be passed
-				 * to the shell.
-				 */
-				/* rewind the (not yet handled) option */
-				optind--;
-				goto end_su_options;
-				break;	/* NOT REACHED */
 			case 'c':
 				command = optarg;
 				break;
@@ -374,7 +397,7 @@ int main (int argc, char **argv)
 				usage ();	/* NOT REACHED */
 			}
 		}
-	      end_su_options:
+
 		if (optind < argc && !strcmp (argv[optind], "-")) {
 			fakelogin = 1;
 			optind++;
@@ -447,7 +470,7 @@ int main (int argc, char **argv)
 	 * Sort out the password of user calling su, in case needed later
 	 * -- chris
 	 */
-	if ((spwd = getspnam (oldname)))
+	if ((spwd = getspnam (oldname))) /* !USE_PAM, no need for xgetspnam */
 		pw->pw_passwd = spwd->sp_pwdp;
 	oldpass = xstrdup (pw->pw_passwd);
 #endif				/* SU_ACCESS */
@@ -482,7 +505,7 @@ int main (int argc, char **argv)
 	 * The password file entries for the user is gotten and the account
 	 * validated.
 	 */
-	if (!(pw = getpwnam (name))) {
+	if (!(pw = xgetpwnam (name))) {
 		(void) fprintf (stderr, _("Unknown id: %s\n"), name);
 		closelog ();
 		exit (1);
@@ -490,7 +513,7 @@ int main (int argc, char **argv)
 #ifndef USE_PAM
 	spwd = NULL;
 	if (strcmp (pw->pw_passwd, SHADOW_PASSWD_STRING) == 0
-	    && (spwd = getspnam (name)))
+	    && (spwd = getspnam (name))) /* !USE_PAM, no need for xgetspnam */
 		pw->pw_passwd = spwd->sp_pwdp;
 #endif				/* !USE_PAM */
 	pwent = *pw;
@@ -671,8 +694,10 @@ int main (int argc, char **argv)
 			spwd = pwd_to_spwd (&pwent);
 
 		if (expire (&pwent, spwd)) {
+			/* !USE_PAM, no need for xgetpwnam */
 			struct passwd *pwd = getpwnam (name);
 
+			/* !USE_PAM, no need for xgetspnam */
 			spwd = getspnam (name);
 			if (pwd)
 				pwent = *pwd;
@@ -707,11 +732,6 @@ int main (int argc, char **argv)
 		addenv ("PATH", cp);
 	}
 
-#ifndef USE_PAM
-	/* setup the environment for PAM later on, else we run into auth problems */
-	environ = newenvp;	/* make new environment active */
-#endif
-
 	if (getenv ("IFS"))	/* don't export user IFS ... */
 		addenv ("IFS= \t\n", NULL);	/* ... instead, set a safe IFS */
 
@@ -761,7 +781,7 @@ int main (int argc, char **argv)
 		SYSLOG ((LOG_ERR, "pam_open_session: %s",
 			 pam_strerror (pamh, ret)));
 		fprintf (stderr, _("%s: %s\n"), Prog, pam_strerror (pamh, ret));
-		pam_setcred(pamh, PAM_DELETE_CRED);
+		pam_setcred (pamh, PAM_DELETE_CRED);
 		pam_end (pamh, ret);
 		exit (1);
 	}
@@ -785,13 +805,16 @@ int main (int argc, char **argv)
 
 	/* become the new user */
 	if (change_uid (&pwent)) {
-		pam_close_session(pamh, 0);
+		pam_close_session (pamh, 0);
 		pam_setcred (pamh, PAM_DELETE_CRED);
 		pam_end (pamh, PAM_ABORT);
 		exit (1);
 	}
 #else				/* !USE_PAM */
-	if (!amroot)		/* no limits if su from root */
+	environ = newenvp;	/* make new environment active */
+
+	/* no limits if su from root (unless su must fake login's behavior) */
+	if (!amroot || fakelogin)
 		setup_limits (&pwent);
 
 	if (setup_uid_gid (&pwent, is_console))
@@ -853,7 +876,7 @@ int main (int argc, char **argv)
 #ifndef USE_PAM
 		(void) execve (shellstr, &argv[-1], environ);
 		err = errno;
-		(void) fprintf (stderr, _("No shell\n"));
+		(void) fputs (_("No shell\n"), stderr);
 		SYSLOG ((LOG_WARN, "Cannot execute %s", shellstr));
 		closelog ();
 		exit (err == ENOENT ? E_CMD_NOTFOUND : E_CMD_NOEXEC);
